module EIO(readRawFile,writeRawFile, makeBitHandle, getBits, putBits, BitHandle, endBitHandle) where

import Control.Exception
import Word
import IO
import Data.Array.IO
import IOExts
import Bits
import Char
import GenUtil

bufSize = 1024

readRawFile :: String -> IO [Word8]
readRawFile fn = Control.Exception.bracket (openFile fn ReadMode) hClose hGetRawContents 

writeRawFile :: String -> [Word8] -> IO ()
writeRawFile fn xs = Control.Exception.bracket (openFile fn WriteMode) hClose $ \h -> hPutRawContents h xs

hGetRawContents :: Handle -> IO [Word8]
hGetRawContents h = do
    a <- newArray_ (0,bufSize)
    getall a where
	getall a = do
	    sz <- hGetArray h a bufSize
	    av <- getElems a
	    if sz == 0 then return [] else do
		r <- getall a
		return (take sz av ++ r)
    
hPutRawContents :: Handle -> [Word8] -> IO ()
hPutRawContents h xs = do
    a <- newArray_ (0,bufSize)
    prc a h xs where
	prc _ _ [] = return ()
	prc a h xs@(_:_) = do
	    let (ys,zs) = splitAt bufSize xs
	    let lys = length ys
	    mapM_ (\(i,e) -> writeArray a i e) (zip [0..lys - 1] ys)
	    hPutArray h a lys
	    prc a h zs


type BitHandle = (Handle, IORef (Word32,Int))



makeBitHandle :: Handle -> IO BitHandle
makeBitHandle h = do
    r <- newIORef (0,0)
    return (h,r)

endBitHandle :: BitHandle -> IO ()
endBitHandle (h,r) = do
    (w8,c) <- readIORef r 
    case c of 
	0 -> return ()
	n -> hPutChar h (chr $ fromIntegral w8)

putBits :: BitHandle -> Int -> Word32 -> IO ()
putBits _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBits _ 0 _ = return ()
putBits bh@(h,r) nbits val = do
    putBit (val `mod` 2)
    putBits bh (nbits - 1) (val `shiftR` 1) where
	putBit b = do
	    (w8,c) <- readIORef r
	    let w8' = w8 .|. (b `shiftL` c)
	    case c of 
		7 -> hPutChar h (chr $ fromIntegral w8') >> writeIORef r (0,0)
		_ -> writeIORef r (w8',c + 1)

getBits :: BitHandle -> Int -> IO Word32
getBits _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBits _ 0 = return 0
getBits bh@(h,r) nbits = do
    xs <- replicateM nbits getBit
    let cs = reverse xs 
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 cs where
	getBit = do
	    (w8,c) <- readIORef r
	    if c > 0 then 
		writeIORef r (w8 `shiftR` 1, c - 1) >> return (w8 .&. 1)
	     else do
		w8 <- fmap (fromIntegral . ord) $ hGetChar h
		writeIORef r (w8,8)
		getBit

{-
putBitsRaw :: Handle -> Int -> Word32 -> IO ()
putBitsRaw _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBitsRaw _ 0 _ = return ()
putBitsRaw bh nbits val = do
    hPutChar bh ("01" !! ((fromIntegral val) `mod` 2))
    putBitsRaw bh (nbits - 1) (val `shiftR` 1)

getBitsRaw :: Handle -> Int -> IO Word32
getBitsRaw _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBitsRaw _ 0 = return 0
getBitsRaw bh nbits = do
    xs <- replicateM nbits (hGetChar bh)
    let cs = map f $ reverse xs 
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 cs where
	f '0' = 0
	f '1' = 1
	f x = error $ "invalid character in savefile: " ++ show x
-}
